@@ -27,10 +27,16 @@ type Engine struct {
 	runArgs   []string
 	running   atomic.Bool
 
-	eventCh        chan string
-	watcherStopCh  chan bool
-	buildRunCh     chan bool
-	buildRunStopCh chan bool
+	eventCh       chan string
+	watcherStopCh chan bool
+	// buildRunCh serves dual purpose:
+	// 1. As a semaphore ensuring only one build runs at a time (buffer size 1)
+	// 2. Carries each build's unique stop channel for cancellation
+	// When a new build starts, it retrieves the previous build's stop channel,
+	// closes it to signal cancellation, then inserts its own fresh channel.
+	// This prevents the race condition where a new build could consume a stop
+	// signal meant for a previous build (issue #784).
+	buildRunCh chan chan struct{}
 	// binStopCh is a channel for process termination control
 	// Type chan<- chan int indicates it's a send-only channel that transmits another channel(chan int)
 	binStopCh chan<- chan int
@@ -60,20 +66,19 @@ func NewEngineWithConfig(cfg *Config, debugMode bool) (*Engine, error) {
 	}
 	runArgs = append(runArgs, cfg.Build.ArgsBin...)
 	e := Engine{
-		config:         cfg,
-		exiter:         defaultExiter{},
-		proxy:          NewProxy(&cfg.Proxy),
-		logger:         logger,
-		watcher:        watcher,
-		debugMode:      debugMode,
-		runArgs:        runArgs,
-		eventCh:        make(chan string, 1000),
-		watcherStopCh:  make(chan bool, 10),
-		buildRunCh:     make(chan bool, 1),
-		buildRunStopCh: make(chan bool, 1),
-		exitCh:         make(chan bool),
-		fileChecksums:  &checksumMap{m: make(map[string]string)},
-		watchers:       0,
+		config:        cfg,
+		exiter:        defaultExiter{},
+		proxy:         NewProxy(&cfg.Proxy),
+		logger:        logger,
+		watcher:       watcher,
+		debugMode:     debugMode,
+		runArgs:       runArgs,
+		eventCh:       make(chan string, 1000),
+		watcherStopCh: make(chan bool, 10),
+		buildRunCh:    make(chan chan struct{}, 1),
+		exitCh:        make(chan bool),
+		fileChecksums: &checksumMap{m: make(map[string]string)},
+		watchers:      0,
 	}
 
 	return &e, nil
@@ -405,11 +410,13 @@ func (e *Engine) start() {
 			// go down
 		}
 
-		// already build and run now
+		// Stop any currently running build by closing its stop channel
 		select {
-		case <-e.buildRunCh:
-			e.buildRunStopCh <- true
+		case oldStopCh := <-e.buildRunCh:
+			// Close the old build's stop channel to signal it to stop
+			close(oldStopCh)
 		default:
+			// No build is currently running
 		}
 
 		// if current app is running, stop it
@@ -420,16 +427,25 @@ func (e *Engine) start() {
 }
 
 func (e *Engine) buildRun() {
-	e.buildRunCh <- true
+	// Create this build's unique stop channel
+	myStopCh := make(chan struct{})
+
+	// Put our stop channel in buildRunCh (acts as semaphore + carries our stop token)
+	e.buildRunCh <- myStopCh
 	defer func() {
 		<-e.buildRunCh
 	}()
 
+	// Check if we were already signaled to stop before we even started
 	select {
-	case <-e.buildRunStopCh:
+	case <-myStopCh:
+		return
+	case <-e.exitCh:
+		e.mainDebug("exit in buildRun before pre_cmd")
 		return
 	default:
 	}
+
 	var err error
 	if err = e.runPreCmd(); err != nil {
 		e.runnerLog("failed to execute pre_cmd: %s", err.Error())
@@ -455,14 +471,16 @@ func (e *Engine) buildRun() {
 		}
 	}
 
+	// Check again before running the binary
 	select {
-	case <-e.buildRunStopCh:
+	case <-myStopCh:
 		return
 	case <-e.exitCh:
-		e.mainDebug("exit in buildRun")
+		e.mainDebug("exit in buildRun after build")
 		return
 	default:
 	}
+
 	if err = e.runBin(); err != nil {
 		e.runnerLog("failed to run, error: %s", err.Error())
 	}
@@ -648,6 +666,7 @@ func (e *Engine) runBin() error {
 				processExit := make(chan struct{})
 				e.mainDebug("running process pid %v", cmd.Process.Pid)
 				if e.config.Proxy.Enabled {
+					e.mainDebug("reloading proxy")
 					e.proxy.Reload()
 				}
 
