@@ -21,6 +21,8 @@ import (
 
 const (
 	sliceCmdArgSeparator = ","
+	// extWildcard is used in include_ext to match all file extensions
+	extWildcard = "*"
 )
 
 func (e *Engine) mainLog(format string, v ...interface{}) {
@@ -172,6 +174,10 @@ func (e *Engine) checkIncludeFile(path string) bool {
 func (e *Engine) isIncludeExt(path string) bool {
 	ext := filepath.Ext(path)
 	for _, v := range e.config.Build.IncludeExt {
+		if strings.TrimSpace(v) == extWildcard {
+			// Wildcard matches all files, but exclude the binary file
+			return !e.isBinPath(path)
+		}
 		if ext == "."+strings.TrimSpace(v) {
 			return true
 		}
@@ -179,6 +185,24 @@ func (e *Engine) isIncludeExt(path string) bool {
 	return false
 }
 
+// isBinPath checks if the given path is the binary file path
+func (e *Engine) isBinPath(path string) bool {
+	binPath := e.config.binPath()
+	if binPath == "" {
+		return false
+	}
+	// Normalize the path for comparison
+	absPath, err := filepath.Abs(path)
+	if err != nil {
+		return false
+	}
+	absBinPath, err := filepath.Abs(binPath)
+	if err != nil {
+		return false
+	}
+	return absPath == absBinPath
+}
+
 func (e *Engine) isExcludeRegex(path string) (bool, error) {
 	regexes, err := e.config.Build.RegexCompiled()
 	if err != nil {
