@@ -1,7 +1,6 @@
 package runner
 
 import (
-	"errors"
 	"fmt"
 	"os"
 	"os/exec"
@@ -10,7 +9,6 @@ import (
 	"runtime"
 	"strconv"
 	"strings"
-	"syscall"
 	"testing"
 	"time"
 
@@ -58,6 +56,7 @@ func TestExpandPathWithHomePath(t *testing.T) {
 }
 
 func TestNormalizeIncludeDirOutsideRoot(t *testing.T) {
+	t.Parallel()
 	root := t.TempDir()
 	parent := filepath.Dir(root)
 	external := filepath.Join(parent, "pkg")
@@ -80,6 +79,7 @@ func TestNormalizeIncludeDirOutsideRoot(t *testing.T) {
 }
 
 func TestCheckIncludeDirRestrictsWithinRoot(t *testing.T) {
+	t.Parallel()
 	root := t.TempDir()
 	runnerDir := filepath.Join(root, "runner")
 	require.NoError(t, os.Mkdir(runnerDir, 0o755))
@@ -105,6 +105,7 @@ func TestCheckIncludeDirRestrictsWithinRoot(t *testing.T) {
 }
 
 func TestFileChecksum(t *testing.T) {
+	t.Parallel()
 	tests := []struct {
 		name                  string
 		fileContents          []byte
@@ -165,6 +166,7 @@ func TestFileChecksum(t *testing.T) {
 }
 
 func TestChecksumMap(t *testing.T) {
+	t.Parallel()
 	m := &checksumMap{m: make(map[string]string, 3)}
 
 	if !m.updateFileChecksum("foo.txt", "abcxyz") {
@@ -185,6 +187,7 @@ func TestChecksumMap(t *testing.T) {
 }
 
 func TestAdaptToVariousPlatforms(t *testing.T) {
+	t.Parallel()
 	config := &Config{
 		Build: cfgBuild{
 			Bin: "tmp\\main.exe  -dev",
@@ -196,27 +199,6 @@ func TestAdaptToVariousPlatforms(t *testing.T) {
 	}
 }
 
-func Test_killCmd_no_process(t *testing.T) {
-	e := Engine{
-		config: &Config{
-			Build: cfgBuild{
-				SendInterrupt: false,
-			},
-		},
-	}
-	_, err := e.killCmd(&exec.Cmd{
-		Process: &os.Process{
-			Pid: 9999,
-		},
-	})
-	if err == nil {
-		t.Errorf("expected error but got none")
-	}
-	if !errors.Is(err, syscall.ESRCH) {
-		t.Errorf("expected '%s' but got '%s'", syscall.ESRCH, errors.Unwrap(err))
-	}
-}
-
 func Test_killCmd_SendInterrupt_false(t *testing.T) {
 	_, b, _, _ := runtime.Caller(0)
 
@@ -345,6 +327,7 @@ func Test_killCmd_KillsDetachedChildren(t *testing.T) {
 }
 
 func TestGetStructureFieldTagMap(t *testing.T) {
+	t.Parallel()
 	c := Config{}
 	tagMap := flatConfig(c)
 	assert.NotEmpty(t, tagMap)
@@ -354,6 +337,7 @@ func TestGetStructureFieldTagMap(t *testing.T) {
 }
 
 func TestSetStructValue(t *testing.T) {
+	t.Parallel()
 	c := Config{}
 	v := reflect.ValueOf(&c)
 	setValue2Struct(v, "TmpDir", "asdasd")
@@ -361,6 +345,7 @@ func TestSetStructValue(t *testing.T) {
 }
 
 func TestNestStructValue(t *testing.T) {
+	t.Parallel()
 	c := Config{}
 	v := reflect.ValueOf(&c)
 	setValue2Struct(v, "Build.Cmd", "asdasd")
@@ -368,6 +353,7 @@ func TestNestStructValue(t *testing.T) {
 }
 
 func TestNestStructArrayValue(t *testing.T) {
+	t.Parallel()
 	c := Config{}
 	v := reflect.ValueOf(&c)
 	setValue2Struct(v, "Build.ExcludeDir", "dir1,dir2")
@@ -375,6 +361,7 @@ func TestNestStructArrayValue(t *testing.T) {
 }
 
 func TestNestStructArrayValueOverride(t *testing.T) {
+	t.Parallel()
 	c := Config{
 		Build: cfgBuild{
 			ExcludeDir: []string{"default1", "default2"},
@@ -386,6 +373,7 @@ func TestNestStructArrayValueOverride(t *testing.T) {
 }
 
 func TestCheckIncludeFile(t *testing.T) {
+	t.Parallel()
 	e := Engine{
 		config: &Config{
 			Build: cfgBuild{
@@ -398,7 +386,95 @@ func TestCheckIncludeFile(t *testing.T) {
 	assert.False(t, e.checkIncludeFile("."))
 }
 
+func TestIsIncludeExt(t *testing.T) {
+	e := Engine{
+		config: &Config{
+			Build: cfgBuild{
+				IncludeExt: []string{"go", "html"},
+			},
+		},
+	}
+	assert.True(t, e.isIncludeExt("main.go"))
+	assert.True(t, e.isIncludeExt("/path/to/file.html"))
+	assert.False(t, e.isIncludeExt("main.js"))
+	assert.False(t, e.isIncludeExt("file"))
+}
+
+func TestIsIncludeExtWildcard(t *testing.T) {
+	tmpDir := t.TempDir()
+	binPath := filepath.Join(tmpDir, "tmp", "main")
+
+	e := Engine{
+		config: &Config{
+			Root: tmpDir,
+			Build: cfgBuild{
+				IncludeExt: []string{"*"},
+				Entrypoint: entrypoint{binPath},
+			},
+		},
+	}
+	// Wildcard should match all file extensions
+	assert.True(t, e.isIncludeExt("main.go"))
+	assert.True(t, e.isIncludeExt("/path/to/file.html"))
+	assert.True(t, e.isIncludeExt("main.js"))
+	assert.True(t, e.isIncludeExt("file.css"))
+	assert.True(t, e.isIncludeExt("file"))           // files without extension
+	assert.True(t, e.isIncludeExt("/path/noext"))    // files without extension
+	assert.False(t, e.isIncludeExt(binPath))         // binary file should be excluded
+	assert.True(t, e.isIncludeExt("some/other/bin")) // other files without extension are ok
+}
+
+func TestIsIncludeExtWildcardWithSpaces(t *testing.T) {
+	e := Engine{
+		config: &Config{
+			Build: cfgBuild{
+				IncludeExt: []string{" * "},
+				Entrypoint: entrypoint{"/tmp/main"},
+			},
+		},
+	}
+	// Wildcard with spaces should still work
+	assert.True(t, e.isIncludeExt("main.go"))
+	assert.True(t, e.isIncludeExt("file.html"))
+}
+
+func TestIsBinPath(t *testing.T) {
+	tmpDir := t.TempDir()
+	binPath := filepath.Join(tmpDir, "tmp", "main")
+
+	e := Engine{
+		config: &Config{
+			Root: tmpDir,
+			Build: cfgBuild{
+				Entrypoint: entrypoint{binPath},
+			},
+		},
+	}
+
+	// Test matching path returns true
+	assert.True(t, e.isBinPath(binPath))
+	// Test non-matching paths return false
+	assert.False(t, e.isBinPath(filepath.Join(tmpDir, "other", "file")))
+	assert.False(t, e.isBinPath("unrelated.go"))
+}
+
+func TestIsBinPathEmptyBinPath(t *testing.T) {
+	// Test when binPath is empty (no entrypoint configured)
+	e := Engine{
+		config: &Config{
+			Build: cfgBuild{
+				Entrypoint: entrypoint{}, // empty entrypoint
+			},
+		},
+	}
+
+	// Should return false when binPath is empty
+	assert.False(t, e.isBinPath("/some/path"))
+	assert.False(t, e.isBinPath("main.go"))
+}
+
 func TestJoinPathRelative(t *testing.T) {
+	t.Parallel()
 	root, err := filepath.Abs("test")
 
 	if err != nil {
@@ -429,6 +505,7 @@ func TestJoinPathAbsolute(t *testing.T) {
 }
 
 func TestFormatPath(t *testing.T) {
+	t.Parallel()
 	type testCase struct {
 		name     string
 		path     string
@@ -511,3 +588,154 @@ func TestFormatPath(t *testing.T) {
 		runTests(t, tests)
 	})
 }
+
+// Test_killCmd_SendInterrupt_FastGracefulExit is a regression test for issue #671.
+// It verifies that when a process exits quickly after receiving SIGINT, Air detects
+// this and returns immediately instead of waiting the full kill_delay.
+//
+// This optimization was implemented in commit 4d26204 (2024-11-01) by Isak Styf.
+// Before the fix, Air would always sleep for the full kill_delay (wasting time).
+// After the fix, Air uses goroutines to detect process exit and returns early.
+//
+// Related: https://github.com/air-verse/air/issues/671
+func Test_killCmd_SendInterrupt_FastGracefulExit(t *testing.T) {
+	if runtime.GOOS == "windows" {
+		t.Skip("send_interrupt not supported on windows")
+	}
+
+	e := Engine{
+		config: &Config{
+			Build: cfgBuild{
+				SendInterrupt: true,
+				KillDelay:     2 * time.Second, // Set high to make the waste observable
+			},
+		},
+	}
+
+	// Process that exits immediately on SIGINT
+	// trap "exit 0" INT means: exit cleanly when receiving SIGINT
+	// sleep 100 means: if no signal, run for 100 seconds
+	cmd, _, _, err := e.startCmd(`sh -c 'trap "exit 0" INT; sleep 100'`)
+	require.NoError(t, err, "failed to start command")
+
+	// Give the process a moment to start and set up the signal handler
+	time.Sleep(100 * time.Millisecond)
+
+	start := time.Now()
+	pid, err := e.killCmd(cmd)
+	elapsed := time.Since(start)
+
+	require.NoError(t, err, "killCmd should succeed")
+	assert.Positive(t, pid, "should return valid pid")
+
+	// Core assertion: should complete in much less than 2 seconds
+	// Process exits immediately on SIGINT, so should finish in <500ms
+	// With the fix (commit 4d26204), this should PASS
+	// Without the fix, this would FAIL (would take 2+ seconds)
+	assert.Less(t, elapsed, 500*time.Millisecond,
+		"killCmd should return quickly when process exits gracefully on SIGINT, "+
+			"but took %v (expected < 500ms). Regression of issue #671!",
+		elapsed)
+
+	t.Logf("✅ PASS: Process exited gracefully in %v (kill_delay was 2s, saved ~%.1fs)",
+		elapsed, 2.0-elapsed.Seconds())
+}
+
+// Test_killCmd_SendInterrupt_IgnoresSIGINT is a regression test for issue #671.
+// It verifies that processes which ignore SIGINT are still killed with SIGKILL
+// after kill_delay. This ensures the optimization (commit 4d26204) doesn't break
+// the fallback behavior for misbehaving processes.
+//
+// Related: https://github.com/air-verse/air/issues/671
+func Test_killCmd_SendInterrupt_IgnoresSIGINT(t *testing.T) {
+	if runtime.GOOS == "windows" {
+		t.Skip("send_interrupt not supported on windows")
+	}
+
+	e := Engine{
+		config: &Config{
+			Build: cfgBuild{
+				SendInterrupt: true,
+				KillDelay:     500 * time.Millisecond,
+			},
+		},
+	}
+
+	// Process that ignores SIGINT
+	// trap "" INT means: ignore SIGINT signal
+	cmd, _, _, err := e.startCmd(`sh -c 'trap "" INT; sleep 100'`)
+	require.NoError(t, err, "failed to start command")
+
+	// Give the process a moment to start and set up the signal handler
+	time.Sleep(100 * time.Millisecond)
+
+	start := time.Now()
+	pid, err := e.killCmd(cmd)
+	elapsed := time.Since(start)
+
+	require.NoError(t, err, "killCmd should succeed")
+	assert.Positive(t, pid, "should return valid pid")
+
+	// Should wait at least kill_delay before sending SIGKILL
+	assert.GreaterOrEqual(t, elapsed, 500*time.Millisecond,
+		"killCmd should wait at least kill_delay when process ignores SIGINT")
+
+	// But should not wait too long after SIGKILL
+	assert.Less(t, elapsed, 1*time.Second,
+		"killCmd should not wait too long after sending SIGKILL, "+
+			"but took %v", elapsed)
+
+	t.Logf("✅ PASS: Process ignored SIGINT, killed with SIGKILL after %v (expected behavior)", elapsed)
+}
+
+// Test_killCmd_SendInterrupt_SlowGracefulExit is a regression test for issue #671.
+// It verifies that when a process takes some time to clean up after receiving
+// SIGINT but still exits within kill_delay, Air returns as soon as the process exits
+// (not waiting the full kill_delay).
+//
+// This optimization was implemented in commit 4d26204 (2024-11-01).
+// Related: https://github.com/air-verse/air/issues/671
+func Test_killCmd_SendInterrupt_SlowGracefulExit(t *testing.T) {
+	if runtime.GOOS == "windows" {
+		t.Skip("send_interrupt not supported on windows")
+	}
+
+	e := Engine{
+		config: &Config{
+			Build: cfgBuild{
+				SendInterrupt: true,
+				KillDelay:     1 * time.Second,
+			},
+		},
+	}
+
+	// Process that takes 300ms to exit after SIGINT (simulating cleanup)
+	// trap "sleep 0.3; exit 0" INT means: when SIGINT received, sleep 300ms then exit
+	cmd, _, _, err := e.startCmd(`sh -c 'trap "sleep 0.3; exit 0" INT; sleep 100'`)
+	require.NoError(t, err, "failed to start command")
+
+	// Give the process a moment to start and set up the signal handler
+	time.Sleep(100 * time.Millisecond)
+
+	start := time.Now()
+	pid, err := e.killCmd(cmd)
+	elapsed := time.Since(start)
+
+	require.NoError(t, err, "killCmd should succeed")
+	assert.Positive(t, pid, "should return valid pid")
+
+	// Should wait at least for the process cleanup time (~300ms)
+	assert.Greater(t, elapsed, 250*time.Millisecond,
+		"should wait at least for process cleanup time (~300ms)")
+
+	// Should return shortly after process exits (~300-500ms)
+	// With the fix (commit 4d26204), this should PASS
+	// Without the fix, this would FAIL (would take 1+ seconds)
+	assert.Less(t, elapsed, 600*time.Millisecond,
+		"killCmd should return soon after process exits, "+
+			"but took %v (expected ~300-500ms). Regression of issue #671!",
+		elapsed)
+
+	t.Logf("✅ PASS: Process exited gracefully in %v after cleanup (kill_delay was 1s, saved ~%.1fs)",
+		elapsed, 1.0-elapsed.Seconds())
+}
